from typing import Dict

import sarge

from cumulusci.core.config import BaseConfig, ConnectedAppOAuthConfig, ScratchOrgConfig
from cumulusci.core.exceptions import (
    CumulusCIException,
    OrgNotFound,
    ServiceNotConfigured,
    ServiceNotValid,
)
from cumulusci.core.sfdx import sfdx

DEFAULT_CONNECTED_APP_PORT = 7788
DEFAULT_CONNECTED_APP_NAME = "built-in"
DEFAULT_CONNECTED_APP = ConnectedAppOAuthConfig(
    {
        "client_id": "3MVG9i1HRpGLXp.or6OVlWVWyn8DXi9xueKNM4npq_AWh.yqswojK9sE5WY7f.biP0w7bNJIENfXc7JMDZGO1",
        "client_secret": None,
        "callback_url": f"http://localhost:{DEFAULT_CONNECTED_APP_PORT}/callback",
    }
)


class BaseProjectKeychain(BaseConfig):
    encrypted = False

    def __init__(self, project_config, key):
        super(BaseProjectKeychain, self).__init__()
        self.config = {
            "orgs": {},
            "app": None,
            "services": {},
        }
        self.project_config = project_config
        self._default_services = {}
        self.key = key
        self._validate_key()
        self._load_keychain()

    def _load_keychain(self):
        self._load_orgs()
        self._load_scratch_orgs()
        self._load_default_connected_app()
        self._load_services()
        self._load_default_services()

    def _validate_key(self):
        pass

    #######################################
    #               Orgs                  #
    #######################################

    def create_scratch_org(self, org_name, config_name, days=None, set_password=True):
        """Adds/Updates a scratch org config to the keychain from a named config"""
        scratch_config = getattr(self.project_config, f"orgs__scratch__{config_name}")
        if days is not None:
            # Allow override of scratch config's default days
            scratch_config["days"] = days
        else:
            # Use scratch config days or default of 1 day
            scratch_config.setdefault("days", 1)
        scratch_config["set_password"] = bool(set_password)
        scratch_config["scratch"] = True
        scratch_config.setdefault("namespaced", False)
        scratch_config["config_name"] = config_name
        scratch_config[
            "sfdx_alias"
        ] = f"{self.project_config.project__name}__{org_name}"
        org_config = ScratchOrgConfig(
            scratch_config, org_name, keychain=self, global_org=False
        )
        org_config.save()

    def set_org(self, org_config, global_org=False, save=True):
        if isinstance(org_config, ScratchOrgConfig):
            org_config.config["scratch"] = True
        self._set_org(org_config, global_org, save=save)

    def set_default_org(self, name):
        """set the default org for tasks and flows by name"""
        org = self.get_org(name)
        self.unset_default_org()
        org.config["default"] = True
        org.save()
        if org.created:
            sfdx(
                sarge.shell_format(
                    "force:config:set defaultusername={}", org.sfdx_alias
                )
            )

    def unset_default_org(self):
        """unset the default orgs for tasks"""
        for org in self.list_orgs():
            org_config = self.get_org(org)
            if org_config.default:
                del org_config.config["default"]
                org_config.save()
        sfdx("force:config:set defaultusername=")

    # This implementation of get_default_org, set_default_org, and unset_default_org
    # is currently kept for backwards compatibility, but EncryptedFileProjectKeychain
    # now stores the default elsewhere, and EnvironmentProjectKeychain doesn't actually
    # persist across multiple invocations of cci, so we should consider getting rid of this.

    def get_default_org(self):
        """retrieve the name and configuration of the default org"""
        for org in self.list_orgs():
            org_config = self.get_org(org)
            if org_config.default:
                return org, org_config
        return None, None

    def get_org(self, name: str):
        """retrieve an org configuration by name key"""
        org = self._get_org(name)
        if org.keychain:
            assert org.keychain is self
        else:
            org.keychain = self
        return org

    def list_orgs(self):
        """list the orgs configured in the keychain"""
        orgs = list(self.orgs.keys())
        orgs.sort()
        return orgs

    def remove_org(self, name, global_org=None):
        if name in self.orgs.keys():
            self._remove_org(name, global_org)
        self.cleanup_org_cache_dirs()

    def _load_orgs(self):
        pass

    def _load_scratch_orgs(self):
        """Creates all scratch org configs for the project in the keychain if
        a keychain org doesn't already exist"""
        current_orgs = self.list_orgs()
        if not self.project_config.orgs__scratch:
            return
        for config_name in self.project_config.orgs__scratch.keys():
            if config_name in current_orgs:
                # Don't overwrite an existing keychain org
                continue
            self.create_scratch_org(config_name, config_name)

    def _set_org(self, org_config, global_org, save=True):
        self.orgs[org_config.name] = org_config

    def _get_org(self, name):
        if name not in self.orgs:
            raise OrgNotFound(f"Org named {name} was not found in keychain")
        return self.orgs.get(name)

    def _remove_org(self, name, global_org):
        del self.orgs[name]
        self._load_orgs()

    def cleanup_org_cache_dirs(self):
        pass  # pragma: no cover

    #######################################
    #              Services               #
    #######################################

    def get_default_service_name(self, service_type: str):
        """Returns the name of the default service for the given type
        or None if no default is currently set for the given type."""
        try:
            return self._default_services[service_type]
        except KeyError:
            return None

    def set_service(
        self,
        service_type: str,
        alias: str,
        service_config: Dict,
        save: bool = True,
        config_encrypted: bool = False,
    ):
        """Store a ServiceConfig in the keychain
        @service_type - type of service
        @alias - name that the service will be stored under
        @service_config - dict of service attributes
        @save - If true, indicates that the service
        should be saved in some manner (subclasses implement).
        @config_encrypted - Indicates whether or not the config
        is already encrypted (as can be the case when reading
        services back from an encrypted file).
        """
        # we only validate attributes when they aren't encrypted
        # if we are setting a service from an encrypted file then it has already been validated
        if not config_encrypted:
            self._validate_service(service_type, alias, service_config)
        self._set_service(
            service_type,
            alias,
            service_config,
            save=save,
            config_encrypted=config_encrypted,
        )

    def get_service(self, service_type, alias=None):
        """Retrieve a stored ServiceConfig from the keychain.
        If no alias is specified then the default service for
        the given type is returned.

        @param service_type: the service to retrieve e.g. 'github'
        @param alias: the alias of the service
        @returns: ServiceConfig for the requested service
        """
        if not self.project_config.services:
            raise ServiceNotValid(
                "Expecting services to be loaded, but none were found."
            )
        elif service_type not in self.project_config.services:
            raise ServiceNotConfigured(
                f"Service type is not configured: {service_type}"
            )

        if service_type not in self.services:
            self._raise_service_not_configured(service_type)

        if not alias:
            alias = self.get_default_service_name(service_type)
            if not alias:
                raise CumulusCIException(
                    f"No default service currently set for service type: {service_type}"
                )
        service = self._get_service(service_type, alias)

        # transparent migration of github API tokens to new key
        if service_type == "github" and service.password and not service.token:
            service.config["token"] = service.password

        return service

    def list_services(self):
        """list the services configured in the keychain"""
        service_types = list(self.services.keys())
        service_types.sort()

        services = {}
        for s_type in service_types:
            if s_type not in services:
                services[s_type] = []
            names = list(self.services[s_type].keys())
            names.sort()
            for name in names:
                services[s_type].append(name)
        return services

    def _load_services(self):
        pass

    def _load_default_services(self):
        self._default_services["connected_app"] = DEFAULT_CONNECTED_APP_NAME

    def _load_default_connected_app(self):
        """Load the default connected app as a first class service on the keychain."""
        if "connected_app" not in self.config["services"]:
            self.config["services"]["connected_app"] = {}
        self.config["services"]["connected_app"][
            DEFAULT_CONNECTED_APP_NAME
        ] = DEFAULT_CONNECTED_APP

    def _set_service(
        self, service_type, alias, service_config, save=True, config_encrypted=False
    ):
        # The first service of a given service_type automatically becomes the default
        if service_type not in self.services:
            self.services[service_type] = {}
            self._default_services[service_type] = alias

        self.services[service_type][alias] = service_config

    def _get_service(self, service_type, alias):
        try:
            return self.services[service_type][alias]
        except KeyError:
            raise ServiceNotConfigured(
                f"No service of type {service_type} configured with name: {alias}"
            )

    def _validate_service(self, service_type, alias, config):
        if (
            not self.project_config.services
            or service_type not in self.project_config.services
        ):
            raise ServiceNotValid(
                f"Service named {alias} is not valid for this project"
            )

        self._validate_service_alias(service_type, alias)
        self._validate_service_attributes(service_type, config)

    def _validate_service_attributes(self, service_type, service_config):
        """
        Validate that all of the required attributes for a
        given service are present.
        """
        missing_required = []
        attr_key = f"services__{service_type}__attributes"
        for atr, config in list(getattr(self.project_config, attr_key).items()):
            if config.get("required") and not getattr(service_config, atr):
                missing_required.append(atr)

        if missing_required:
            if service_type == "github" and missing_required == ["token"]:
                service_config.token = service_config.password
                return

            raise ServiceNotValid(
                f"Missing required attribute(s) for {service_type} service: {missing_required}"
            )

    def _validate_service_alias(self, service_type, alias):
        if alias == service_type:
            raise ServiceNotValid(
                "Service name cannot be the same as the service type."
            )
        if service_type == "connected_app" and alias == DEFAULT_CONNECTED_APP_NAME:
            raise ServiceNotValid(
                f"You cannot use the name {DEFAULT_CONNECTED_APP_NAME} for a connected app service. Please select a different name."
            )

    def _raise_service_not_configured(self, service_type):
        service_types = ", ".join(list(self.services))
        raise ServiceNotConfigured(
            f"Service type {service_type} is not configured for this project. Configured services are: {service_types}"
        )

    @property
    def cache_dir(self):
        "Helper function to get the cache_dir from the project_config"
        return self.project_config.cache_dir  # pragma: no cover



import base64
import json
import os
import pickle
import sys
import typing as T
from pathlib import Path
from shutil import rmtree

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC

from cumulusci.core.config import OrgConfig, ScratchOrgConfig, ServiceConfig
from cumulusci.core.exceptions import (
    ConfigError,
    CumulusCIException,
    CumulusCIUsageError,
    KeychainKeyNotFound,
    OrgNotFound,
    ServiceNotConfigured,
)
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.keychain.base_project_keychain import DEFAULT_CONNECTED_APP_NAME
from cumulusci.core.utils import import_class, import_global

DEFAULT_SERVICES_FILENAME = "DEFAULT_SERVICES.json"

# The file permissions that we want set on all
# .org and .service files. Equivalent to -rw-------
SERVICE_ORG_FILE_MODE = 0o600
OS_FILE_FLAGS = os.O_WRONLY | os.O_CREAT
if sys.platform.startswith("win"):
    # O_BINARY only available on Windows
    OS_FILE_FLAGS |= os.O_BINARY


BS = 16
backend = default_backend()


def pad(s):
    return s + (BS - len(s) % BS) * chr(BS - len(s) % BS).encode("ascii")


scratch_org_class = os.environ.get("CUMULUSCI_SCRATCH_ORG_CLASS")
if scratch_org_class:
    scratch_org_factory = import_global(scratch_org_class)  # pragma: no cover
else:
    scratch_org_factory = ScratchOrgConfig


"""
EncryptedFileProjectKeychain
----------------------------
This class represents a project keychain that stores services and orgs in encrypted
files. These files have the extensions: '.org' and '.service'.

Service files (.service) are organized under ~/.cumulusci/services/ into sub-directories that
pertain to a particular service type (e.g. github, connected_app). These files hold
the encrypted 'attributes' of each service as they are listed in the universal
cumulusci.yml file. There are multiple DEFAULT_SERVICES.json files that map a particular
service type to the default alias for that taype.

The DEFAULT_SERVICES.json file that resides at ~/.cumulusci holds the global default
services mappings. Each local project directory (~/.cumulusci/project/) also has a
DEFAULT_SERVICES.json file for any services that have been configured to be project specific.

Org files (.org) live in both under ~/.cumulusci and in local project directories.
The .org files store an encrypted access_token and other org attributes. There is a
DEFAULT_ORG.txt file in each local project directory that stores the name of
the default org for that project.
"""


class EncryptedFileProjectKeychain(BaseProjectKeychain):
    encrypted = True
    env_service_alias_prefix = "env"
    env_service_var_prefix = "CUMULUSCI_SERVICE_"
    env_org_var_prefix = "CUMULUSCI_ORG_"

    @property
    def global_config_dir(self):
        try:
            global_config_dir = (
                self.project_config.universal_config_obj.cumulusci_config_dir
            )
        except AttributeError:
            # Handle a global config passed as a project config
            global_config_dir = self.project_config.cumulusci_config_dir
        return global_config_dir

    @property
    def project_local_dir(self):
        return self.project_config.project_local_dir

    #######################################
    #             Encryption              #
    #######################################

    def _get_cipher(self, iv=None):
        key = self.key
        if not isinstance(key, bytes):
            key = key.encode()
        if iv is None:
            iv = os.urandom(16)
        cipher = Cipher(AES(key), CBC(iv), backend=backend)
        return cipher, iv

    def _encrypt_config(self, config):
        pickled = pickle.dumps(config.config, protocol=2)
        pickled = pad(pickled)
        cipher, iv = self._get_cipher()
        return base64.b64encode(iv + cipher.encryptor().update(pickled))

    def _decrypt_config(self, config_class, encrypted_config, extra=None, context=None):
        if self.key:
            if not encrypted_config:
                if extra:
                    return config_class(None, *extra)
                else:
                    return config_class()
            encrypted_config = base64.b64decode(encrypted_config)
            iv = encrypted_config[:16]
            cipher, iv = self._get_cipher(iv)
            pickled = cipher.decryptor().update(encrypted_config[16:])
            try:
                unpickled = pickle.loads(pickled, encoding="bytes")
            except Exception:
                raise KeychainKeyNotFound(
                    f"Unable to decrypt{' ' + context if context else ''}. "
                    "It was probably stored using a different CUMULUSCI_KEY."
                )
            # Convert bytes created in Python 2
            config_dict = {}
            for k, v in unpickled.items():
                if isinstance(k, bytes):
                    k = k.decode("utf-8")
                if isinstance(v, bytes):
                    v = v.decode("utf-8")
                config_dict[k] = v

        args = [config_dict]
        if extra:
            args += extra
        return self._construct_config(config_class, args)

    def _construct_config(self, config_class, args):
        if args[0].get("scratch"):
            config_class = ScratchOrgConfig

        return config_class(*args)

    def _get_config_bytes(self, config) -> bytes:
        """Depending on if a key is present return
        the bytes that we want to store on the keychain."""
        org_bytes = None
        if self.key:
            org_bytes = self._encrypt_config(config)
        else:
            org_bytes = pickle.dumps(config.config)

        assert org_bytes is not None, "org_bytes should have a value"
        return org_bytes

    def _validate_key(self):
        # the key can be None when we detect issues using keyring
        if not self.key:
            return
        if len(self.key) != 16:
            raise ConfigError("The keychain key must be 16 characters long.")

    #######################################
    #               Orgs                  #
    #######################################

    def get_default_org(self):
        """Retrieve the name and configuration of the default org"""
        # first look for a file with the default org in it
        default_org_path = self._default_org_path
        if default_org_path and default_org_path.exists():
            org_name = default_org_path.read_text().strip()
            try:
                org_config = self.get_org(org_name)
                return org_name, org_config
            except OrgNotFound:  # org was deleted
                default_org_path.unlink()  # we don't really have a usable default anymore

        # fallback to old way of doing it
        org_name, org_config = super().get_default_org()
        if org_name:
            self.set_default_org(org_name)  # upgrade to new way
        return org_name, org_config

    def set_default_org(self, name: str):
        """Set the default org for tasks and flows by name"""
        super().set_default_org(name)
        self._default_org_path.write_text(name)

    def unset_default_org(self):
        """Unset the default orgs for tasks and flows"""
        super().unset_default_org()
        if self._default_org_path:
            try:
                self._default_org_path.unlink()
            except FileNotFoundError:
                pass

    @property
    def _default_org_path(self):
        if self.project_local_dir:
            return Path(self.project_local_dir) / "DEFAULT_ORG.txt"

    def _load_orgs(self) -> None:
        self._load_orgs_from_environment()
        self._load_org_files(self.global_config_dir, GlobalOrg)
        self._load_org_files(self.project_local_dir, LocalOrg)

    def _load_orgs_from_environment(self):
        for env_var_name, value in os.environ.items():
            if env_var_name.startswith(self.env_org_var_prefix):
                self._load_org_from_environment(env_var_name, value)

    def _load_org_from_environment(self, env_var_name, value):
        org_config = json.loads(value)
        org_name = env_var_name[len(self.env_org_var_prefix) :].lower()
        if org_config.get("scratch"):
            org_config = scratch_org_factory(
                json.loads(value), org_name, keychain=self, global_org=False
            )
        else:
            org_config = OrgConfig(
                org_config, org_name, keychain=self, global_org=False
            )

        self.set_org(org_config, global_org=False, save=False)

    def _load_org_files(self, dirname: str, constructor=None):
        """Loads .org files in a given directory onto the keychain"""
        if not dirname:
            return
        dir_path = Path(dirname)
        for item in sorted(dir_path.iterdir()):
            if item.suffix == ".org":
                with open(item, "rb") as f:
                    config = f.read()
                name = item.name.replace(".org", "")
                if "orgs" not in self.config:
                    self.config["orgs"] = {}
                self.config["orgs"][name] = (
                    constructor(config) if constructor else config
                )

    def _set_org(self, org_config, global_org, save=True):
        if org_config.keychain:
            assert org_config.keychain is self
        assert org_config.global_org == global_org
        org_config.keychain = self
        org_config.global_org = global_org

        org_name = org_config.name

        org_bytes = self._get_config_bytes(org_config)

        if global_org:
            org_config = GlobalOrg(org_bytes)
        else:
            org_config = LocalOrg(org_bytes)

        self.orgs[org_name] = org_config

        # if keychain is explicitly set to
        # EnvironmentProjectKeychain never save the org config.
        keychain_class = os.environ.get("CUMULUSCI_KEYCHAIN_CLASS")
        if keychain_class == "EnvironmentProjectKeychain":
            message = (
                "The keychain is currently set to EnvironmentProjectKeychain; "
                "skipping save of org config to file. "
                "If you would like orgs to be saved for re-use later, remove the "
                "CUMULUSCI_KEYCHAIN_CLASS environment variable."
            )
            self.logger.warning(message)
            return

        if save:
            self._save_org(
                org_name,
                org_config.data,
                global_org,
            )

    def _save_org(self, name, org_bytes, global_org):
        """
        @name - name of the org
        @org_bytes - bytes-like objecte to write to disk
        @global_org - whether or not this is a global org
        """
        if global_org:
            filename = Path(f"{self.global_config_dir}/{name}.org")
        elif self.project_local_dir is None:
            return
        else:
            filename = Path(f"{self.project_local_dir}/{name}.org")

        fd = os.open(filename, OS_FILE_FLAGS, SERVICE_ORG_FILE_MODE)
        with open(fd, "wb") as f:
            f.write(org_bytes)

    def _get_org(self, name):
        try:
            config = self.orgs[name].data
            global_org = self.orgs[name].global_org
        except KeyError:
            raise OrgNotFound(f"Org with name '{name}' does not exist.")

        if self.key:
            org = self._decrypt_config(
                OrgConfig,
                config,
                extra=[name, self],
                context=f"org config ({name})",
            )
        else:
            config = pickle.loads(config)
            org = self._construct_config(OrgConfig, [config, name, self])

        org.global_org = global_org
        return org

    def _remove_org(self, name, global_org):
        scope = self.global_config_dir if global_org else self.project_local_dir
        org_path = Path(f"{scope}/{name}.org")
        if not org_path.exists():
            if not global_org:
                raise OrgNotFound(
                    f"Could not find org named {name} to delete.  Deleting in project org mode.  Is {name} a global org?"
                )
            raise OrgNotFound(
                f"Could not find org named {name} to delete.  Deleting in global org mode.  Is {name} a project org instead of a global org?"
            )

        org_path.unlink()
        del self.orgs[name]

    def cleanup_org_cache_dirs(self):
        """Cleanup directories that are not associated with a connected/live org."""

        if not self.project_config or not self.project_config.cache_dir:
            return
        active_org_domains = set()
        for org in self.list_orgs():
            org_config = self.get_org(org)
            domain = org_config.get_domain()
            if domain:
                active_org_domains.add(domain)

        assert self.project_config.cache_dir, "Project cache dir does not exist."
        assert self.global_config_dir, "Global config directory does not exist."

        project_org_directories = (self.project_config.cache_dir / "orginfo").glob("*")
        global_org_directories = (self.global_config_dir / "orginfo").glob("*")

        for path in list(project_org_directories) + list(global_org_directories):
            if path.is_dir() and path.name not in active_org_domains:
                rmtree(path)

    #######################################
    #              Services               #
    #######################################

    def set_default_service(
        self, service_type: str, alias: str, project: bool = False, save: bool = True
    ) -> None:
        """Public API for setting a default service e.g. `cci service default`

        @param service_type: the type of service
        @param alias: the name of the service
        @param project: Should this be a project default
        @param save: save the defaults so they are loaded on subsequent executions
        @raises ServiceNotConfigured if service_type or alias are invalid
        """
        self._validate_service_type_and_alias(service_type, alias)
        self._default_services[service_type] = alias
        if save:
            self._save_default_service(service_type, alias, project=project)

    def rename_service(
        self, service_type: str, current_alias: str, new_alias: str
    ) -> None:
        """Public API for renaming a service

        @param service_type type of service being renamed
        @param current_alias the current alias of the service
        @param new_alias the new alias for the service
        @throws: ServiceNotValid if no services of the given type are configured,
        or if no service of the given type has the current_alias
        """
        if (
            service_type == "connected_app"
            and current_alias == DEFAULT_CONNECTED_APP_NAME
        ):
            raise CumulusCIException(
                "You cannot rename the connected app service that is provided by CumulusCI."
            )

        self._validate_service_type_and_alias(service_type, current_alias)
        if new_alias in self.services[service_type]:
            raise CumulusCIUsageError(
                f"A service of type {service_type} already exists with name: {new_alias}"
            )

        self.services[service_type][new_alias] = self.services[service_type][
            current_alias
        ]
        del self.services[service_type][current_alias]

        # rename the corresponding .service file
        current_filepath = Path(
            f"{self.global_config_dir}/services/{service_type}/{current_alias}.service"
        )
        new_filepath = Path(
            f"{self.global_config_dir}/services/{service_type}/{new_alias}.service"
        )
        current_filepath.replace(new_filepath)

        # look through all DEFAULT_SERVICE.json files and
        # change current_alias to new_alias (if present)
        self._rename_alias_in_default_service_file(
            Path(self.global_config_dir, "DEFAULT_SERVICES.json"),
            service_type,
            current_alias,
            new_alias,
        )
        for project_dir in self._iter_local_project_dirs():
            self._rename_alias_in_default_service_file(
                project_dir / "DEFAULT_SERVICES.json",
                service_type,
                current_alias,
                new_alias,
            )

    def remove_service(self, service_type: str, alias: str):
        """Removes the given service from the keychain. If the service
        is the default service, and there is only one other service
        of the same type, that service is set as the new default.

        @param service_type type of the service
        @param alias the name of the service
        @raises ServiceNotConfigured if the service_type or alias are invalid
        """
        if service_type == "connected_app" and alias == DEFAULT_CONNECTED_APP_NAME:
            raise CumulusCIException(
                f"Unable to remove connected app service: {DEFAULT_CONNECTED_APP_NAME}. "
                "This connected app is provided by CumulusCI and cannot be removed."
            )

        self._validate_service_type_and_alias(service_type, alias)
        # remove the loaded service from the keychain
        del self.services[service_type][alias]

        # delete the corresponding .service file
        service_filepath = Path(
            f"{self.global_config_dir}/services/{service_type}/{alias}.service"
        )
        service_filepath.unlink()

        # remove any references from DEFAULT_SERVICES.json files
        self._remove_reference_to_alias(
            Path(self.global_config_dir, "DEFAULT_SERVICES.json"),
            service_type,
            alias,
        )
        for project_dir in self._iter_local_project_dirs():
            self._remove_reference_to_alias(
                project_dir / "DEFAULT_SERVICES.json",
                service_type,
                alias,
            )

        # if set, remove the service as the default
        if alias == self._default_services[service_type]:
            del self._default_services[service_type]
            if len(self.services[service_type].keys()) == 1:
                alias = self.list_services()[service_type][0]
                self.set_default_service(service_type, alias, project=False)

    def _remove_reference_to_alias(
        self, default_services_filepath: Path, service_type: str, alias: str
    ) -> None:
        """Given the path to a DEFAULT_SERVICES.json file, removes any references
        to the given alias if present."""
        default_services = self._read_default_services(default_services_filepath)

        if service_type in default_services and alias == default_services[service_type]:
            del default_services[service_type]

        self._write_default_services(default_services_filepath, default_services)

    def _rename_alias_in_default_service_file(
        self,
        default_service_file_path: Path,
        service_type: str,
        current_alias: str,
        new_alias: str,
    ) -> None:
        """Given the path to a DEFAULT_SERVICES.json file,
        if current_alias is present for the given service_type,
        then rename it to new_alias. Otherwise, do nothing.
        """
        default_services = self._read_default_services(default_service_file_path)

        if (
            service_type not in default_services
            or current_alias != default_services[service_type]
        ):
            return

        default_services[service_type] = new_alias
        self._write_default_services(default_service_file_path, default_services)

    def _load_services(self) -> None:
        """Load services (and migrate old ones if present)"""
        self._load_services_from_environment()

        if not (self.global_config_dir / "services").is_dir():
            self._create_default_service_files()
            self._create_services_dir_structure()
            self._migrate_services()

        self._load_service_files()

    def _set_service(
        self, service_type, alias, service_config, save=True, config_encrypted=False
    ):
        if service_type not in self.services:
            self.services[service_type] = {}
            # set the first service of a given type as the global default
            self._default_services[service_type] = alias
            if save:
                self._save_default_service(service_type, alias, project=False)

        # If the config is already encrypted coming in
        # (like when were setting services after loading from an encrypted file)
        # then we don't need to do anything.
        if self.key and config_encrypted:
            config = service_config
        elif self.key and not config_encrypted:
            config = self._encrypt_config(service_config)
        else:
            config = pickle.dumps(service_config.config)

        self.services[service_type][alias] = config

        if save:
            self._save_encrypted_service(service_type, alias, config)

    def _save_encrypted_service(self, service_type, alias, encrypted):
        """Write out the encrypted service to disk."""
        service_path = Path(
            f"{self.global_config_dir}/services/{service_type}/{alias}.service"
        )
        if not service_path.parent.is_dir():
            service_path.parent.mkdir()

        fd = os.open(service_path, OS_FILE_FLAGS, SERVICE_ORG_FILE_MODE)
        with open(fd, "wb") as f:
            f.write(encrypted)

    def _get_service(self, service_type, alias):
        if service_type == "connected_app" and alias == DEFAULT_CONNECTED_APP_NAME:
            # CumulusCI's default connected app is not encrypted, just return it
            return self.config["services"]["connected_app"][DEFAULT_CONNECTED_APP_NAME]

        ConfigClass = ServiceConfig
        if "class_path" in self.project_config.config["services"][service_type]:
            class_path = self.project_config.config["services"][service_type][
                "class_path"
            ]
            try:
                ConfigClass = import_class(class_path)
            except (AttributeError, ModuleNotFoundError):
                raise CumulusCIException(
                    f"Unrecognized class_path for service: {class_path}"
                )

        try:
            config = self.services[service_type][alias]
        except KeyError:
            raise ServiceNotConfigured(
                f"No service of type {service_type} exists with the name: {alias}"
            )

        if self.key:
            org = self._decrypt_config(
                ConfigClass,
                config,
                extra=[alias, self],
                context=f"service config ({service_type}:{alias})",
            )
        else:
            config = pickle.loads(config)
            org = self._construct_config(ConfigClass, [config, alias, self])

        return org

    def _load_services_from_environment(self):
        """Load any services specified by environment variables"""
        for env_var_name, value in os.environ.items():
            if env_var_name.startswith(self.env_service_var_prefix):
                self._load_service_from_environment(env_var_name, value)

    def _load_service_from_environment(self, env_var_name, value):
        """Given a valid name/value pair, load the
        service from the environment on to the keychain"""
        service_config = ServiceConfig(json.loads(value))
        service_type, service_name = self._get_env_service_type_and_name(env_var_name)
        self.set_service(service_type, service_name, service_config, save=False)

    def _get_env_service_type_and_name(self, env_service_name):
        return (
            self._get_env_service_type(env_service_name),
            self._get_env_service_name(env_service_name),
        )

    def _get_env_service_type(self, env_service_name):
        """Parse the service type given the env var name"""
        post_prefix = env_service_name[len(self.env_service_var_prefix) :].lower()
        return post_prefix.split("__")[0]

    def _get_env_service_name(self, env_service_name):
        """
        Parse the service name given the env var name.
        Services from the environment can be listed with or without a name:
        * CUMULUSCI_SERVICE_service_type -> this gets a default name of "env"
        * CUMULUSCI_SERVICE_service_type__name -> this gets the name "env-name"
        """
        parts = env_service_name.split("__")
        return f"env-{parts[-1]}" if len(parts) > 1 else "env"

    def _load_service_files(self) -> None:
        """
        Load configured services onto the keychain.
        This method recursively goes through all subdirectories
        in ~/.cumulusci/services looking for .service files to load.
        """
        services_dir = Path(f"{self.global_config_dir}/services")
        for item in services_dir.glob("**/*"):
            if item.suffix == ".service":
                with open(item) as f:
                    config = f.read()
                name = item.name.replace(".service", "")
                service_type = item.parent.name

                self.set_service(
                    service_type, name, config, save=False, config_encrypted=True
                )

    def _load_default_services(self) -> None:
        """Init self._default_services on the keychain so that
        calls to get_service() that do not pass an alias can
        return the default service for the given type"""

        # set CumulusCI's default connected app as the default first
        # so it will be overwritten if the user is using a different connected app
        self._default_services["connected_app"] = DEFAULT_CONNECTED_APP_NAME

        global_default_services = Path(
            f"{self.global_config_dir}/{DEFAULT_SERVICES_FILENAME}"
        )
        self._set_default_services_from_dir(global_default_services)

        project_default_services = Path(
            f"{self.project_local_dir}/{DEFAULT_SERVICES_FILENAME}"
        )
        # project defaults will overwrite global defaults
        self._set_default_services_from_dir(project_default_services)

    def _set_default_services_from_dir(self, default_services_file: Path) -> None:
        """Sets the keychain._default_services dictionary to the default
        values in the given file.

        @param default_services_file path to a DEFAULT_SERVICES.json file
        """
        default_services = self._read_default_services(default_services_file)
        for s_type, alias in default_services.items():
            self._default_services[s_type] = alias

    def _save_default_service(
        self, service_type: str, alias: str, project: bool = False
    ) -> None:
        """Write out the contents of self._default_services to the
        DEFAULT_SERVICES.json file based on the provided scope"""
        dir_path = (
            Path(self.project_local_dir) if project else Path(self.global_config_dir)
        )
        default_services_file = dir_path / DEFAULT_SERVICES_FILENAME

        default_services = self._read_default_services(default_services_file)
        default_services[service_type] = alias
        self._write_default_services(default_services_file, default_services)

    def _create_default_service_files(self) -> None:
        """
        Generate the DEFAULT_SERVICES.json files at global and project scopes.

        @param local_proj_path: should be the local_project_path for the project
        """
        global_default_service_file = Path(
            f"{self.global_config_dir}/{DEFAULT_SERVICES_FILENAME}"
        )
        if global_default_service_file.is_file():
            return

        self._write_default_services_for_dir(self.global_config_dir)
        for local_proj_dir in self._iter_local_project_dirs():
            self._write_default_services_for_dir(local_proj_dir)

    def _write_default_services_for_dir(self, dir_path: str) -> None:
        """Look through the given dir and set all .service files
        present as the defaults for their given types, and write these
        out to a DEFAULT_SERVICES.json file in the directory. This occurs
        once before .service files are migrated to the appropriate
        services/ sub-directory, so we set the default to the alias
        that will be assigned during migration.

        @param dir_path: the directory to look through and write
        the DEFAULT_SERVICES.json file in.
        """
        dir_path = Path(dir_path)
        default_services = {}
        for item in dir_path.iterdir():
            if item.suffix == ".service":
                service_type = item.name.replace(".service", "")
                alias = (
                    "global"
                    if item.parent.name == ".cumulusci"
                    else Path(self.project_local_dir).name
                )
                default_services[service_type] = f"{alias}"

        self._write_default_services(
            dir_path / DEFAULT_SERVICES_FILENAME, default_services
        )

    def _create_services_dir_structure(self) -> None:
        """
        Ensure the 'services' directory structure exists.
        The services dir has the following structure and lives
        in the global_config_dir:

        services
        |-- github
        |   |-- alias1.service
        |   |-- alias2.service
        |   |-- ...
        |-- devhub
        |   |-- alias1.service
        |   |-- alias2.service
        |   |-- ...
        .

        This also has the advantage that when a new service type
        is added to cumulusci.yml a new directory for that service type
        will be created.
        """
        services_dir_path = Path(f"{self.global_config_dir}/services")
        services_dir_path.mkdir(exist_ok=True)

        configured_service_types = self.project_config.config["services"].keys()
        for service_type in configured_service_types:
            service_type_dir_path = Path(services_dir_path / service_type)
            if not Path.is_dir(service_type_dir_path):
                Path.mkdir(service_type_dir_path)

    def _migrate_services(self):
        """Migrate .service files from the global_config_dir and
        any project local directories."""
        self._migrate_services_from_dir(self.global_config_dir)
        for local_proj_dir in self._iter_local_project_dirs():
            self._migrate_services_from_dir(local_proj_dir)

    def _migrate_services_from_dir(self, dir_path: str) -> None:
        """Migrate all .service files from the given directory to
        the appropriate service sub-directory and apply the default
        alias. This is intended to be run against either the
        global_config_dir, or a local project directory.

        Default aliases are in the form `service_type__scope`.
        Scope is either the name of the local project directory
        or 'global'; depending on where the .service file is located."""
        dir_path = Path(dir_path)
        for item in dir_path.iterdir():
            if item.suffix == ".service":
                new_service_filepath = self._get_new_service_filepath(item)
                if not new_service_filepath:
                    continue
                if new_service_filepath.is_file():
                    self.logger.warning(
                        f"Skipping migration of {item.name} as a default alias already exists for this service type. "
                    )
                    continue

                item.replace(new_service_filepath)

    def _get_new_service_filepath(self, item: Path) -> Path:
        """Given an old .service filepath, determine the path
        and filename of the new .service file.

        @returns: the Path to the newfile, or None if the service
        is of an unrecognized type.
        """
        service_type = item.name.replace(".service", "")
        configured_service_types = self.project_config.config["services"].keys()
        if service_type not in configured_service_types:
            self.logger.info(f"Skipping migration of unrecognized service: {item.name}")
            return None

        alias = (
            "global"
            if item.parent.name == ".cumulusci"
            else Path(self.project_local_dir).name
        )
        new_filename = f"{alias}.service"
        services_sub_dir = Path(f"{self.global_config_dir}/services/{service_type}")
        return services_sub_dir / new_filename

    def _read_default_services(self, file_path: Path) -> T.Dict[str, str]:
        """Reads the default services file at the given path

        @param file_path path to DEFAULT_SERVICES.json
        @returns dict of default services
        @raises CumulusCIException if the file does not exist
        """
        if not file_path.is_file() or file_path.name != DEFAULT_SERVICES_FILENAME:
            return {}
        else:
            return json.loads(file_path.read_text(encoding="utf-8"))

    def _write_default_services(
        self, file_path: Path, default_services: T.Dict[str, str]
    ) -> None:
        """Writes default services out to the given file

        @param file_path path to DEFAULT_SERVICES.json
        @param dictionary mapping service types to the alias of the default service for that type
        @raises CumulusCIException if the file does not exist
        """
        if file_path.name != DEFAULT_SERVICES_FILENAME:
            raise CumulusCIException(
                f"No {DEFAULT_SERVICES_FILENAME} file found at: {file_path}"
            )

        with open(file_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(default_services))

    def _iter_local_project_dirs(self):
        """Iterate over all local project dirs in ~/.cumulusci"""
        for item in Path(self.global_config_dir).iterdir():
            if item.is_dir() and item.name not in ["logs", "services"]:
                yield item

    def _validate_service_type_and_alias(self, service_type, alias):
        """Raises ServiceNotConfigured exception if the service_type
        or alias are not valid."""
        if service_type not in self.services:
            raise ServiceNotConfigured(
                f"No services of type {service_type} are currently configured"
            )
        elif alias not in self.services[service_type]:
            raise ServiceNotConfigured(
                f"No service of type {service_type} configured with the name: {alias}"
            )

    def _raise_service_not_configured(self, name):
        raise ServiceNotConfigured(
            f"'{name}' service configuration could not be found. "
            f"Maybe you need to run: cci service connect {name}"
        )


class GlobalOrg(T.NamedTuple):
    data: bytes
    global_org: bool = True


class LocalOrg(T.NamedTuple):
    data: bytes
    global_org: bool = False



from typing import List, Optional, Tuple
from unittest import mock

import pytest
from github3.exceptions import NotFoundError

from cumulusci.core.config import UniversalConfig
from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.dependencies.dependencies import (
    DynamicDependency,
    GitHubDynamicDependency,
    PackageNamespaceVersionDependency,
    PackageVersionIdDependency,
    StaticDependency,
    UnmanagedGitHubRefDependency,
)
from cumulusci.core.dependencies.resolvers import (
    DependencyResolutionStrategy,
    GitHubBetaReleaseTagResolver,
    GitHubReleaseBranchCommitStatusResolver,
    GitHubReleaseBranchExactMatchCommitStatusResolver,
    GitHubReleaseBranchResolver,
    GitHubReleaseTagResolver,
    GitHubTagResolver,
    GitHubUnmanagedHeadResolver,
    dependency_filter_ignore_deps,
    get_resolver,
    get_resolver_stack,
    get_static_dependencies,
)
from cumulusci.core.exceptions import CumulusCIException, DependencyResolutionError


class ConcreteDynamicDependency(DynamicDependency):
    ref: Optional[str]
    resolved: Optional[bool] = False

    @property
    def is_resolved(self):
        return self.resolved

    def resolve(
        self, context: BaseProjectConfig, strategies: List[DependencyResolutionStrategy]
    ):
        super().resolve(context, strategies)
        self.resolved = True

    @property
    def name(self):
        return ""

    @property
    def description(self):
        return ""


class TestGitHubTagResolver:
    def test_github_tag_resolver(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/ReleasesRepo",
            tag="release/1.0",  # Not the most recent release
        )
        resolver = GitHubTagResolver()

        assert resolver.can_resolve(dep, project_config)
        assert resolver.resolve(dep, project_config) == (
            "tag_sha",
            PackageNamespaceVersionDependency(
                namespace="ccitestdep", version="1.0", package_name="CumulusCI-Test-Dep"
            ),
        )

    def test_github_tag_resolver__2gp(self, project_config):
        tag = mock.Mock()
        tag.return_value.object.sha = "tag_sha"
        tag.return_value.message = """
package_type: 2GP

version_id: 04t000000000000"""
        project_config.get_repo_from_url(
            "https://github.com/SFDO-Tooling/ReleasesRepo"
        ).tag = tag

        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/ReleasesRepo",
            tag="release/1.0",  # Not the most recent release
        )
        resolver = GitHubTagResolver()

        assert resolver.can_resolve(dep, project_config)
        assert resolver.resolve(dep, project_config) == (
            "tag_sha",
            PackageVersionIdDependency(
                version_id="04t000000000000",
                version_number="1.0",
                package_name="CumulusCI-Test-Dep",
            ),
        )

    def test_github_tag_resolver__unmanaged(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/ReleasesRepo",
            tag="release/2.0",
            unmanaged=True,
        )
        resolver = GitHubTagResolver()

        assert resolver.resolve(dep, project_config) == (
            "tag_sha",
            None,
        )

    def test_no_managed_release(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/UnmanagedRepo",  # This repo has no namespace
            tag="release/1.0",
            unmanaged=False,
        )
        resolver = GitHubTagResolver()

        assert resolver.resolve(dep, project_config) == (
            "tag_sha",
            None,
        )

    def test_can_resolve_negative(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/ReleasesRepo"
        )
        resolver = GitHubTagResolver()

        assert not resolver.can_resolve(dep, project_config)

    def test_exception_no_tag_found(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/ReleasesRepo",
            tag="release/3.0",
        )
        resolver = GitHubTagResolver()

        with pytest.raises(DependencyResolutionError) as e:
            resolver.resolve(dep, project_config)
        assert "No release found for tag" in str(e.value)

    def test_str(self):
        assert str(GitHubTagResolver()) == GitHubTagResolver.name


class TestGitHubReleaseTagResolver:
    def test_github_release_tag_resolver(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/ReleasesRepo"
        )
        resolver = GitHubReleaseTagResolver()

        assert resolver.can_resolve(dep, project_config)
        assert resolver.resolve(dep, project_config) == (
            "tag_sha",
            PackageNamespaceVersionDependency(
                namespace="ccitestdep", version="2.0", package_name="CumulusCI-Test-Dep"
            ),
        )

    def test_github_release_tag_resolver__2gp(self, project_config):
        tag = mock.Mock()
        tag.return_value.object.sha = "tag_sha"
        tag.return_value.message = """
package_type: 2GP

version_id: 04t000000000000"""
        project_config.get_repo_from_url(
            "https://github.com/SFDO-Tooling/TwoGPRepo"
        ).tag = tag

        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPRepo"
        )
        resolver = GitHubReleaseTagResolver()

        assert resolver.can_resolve(dep, project_config)
        assert resolver.resolve(dep, project_config) == (
            "tag_sha",
            PackageVersionIdDependency(
                version_id="04t000000000000",
                version_number="1.0",
                package_name="CumulusCI-2GP-Test",
            ),
        )

    def test_beta_release_tag(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/ReleasesRepo"
        )
        resolver = GitHubBetaReleaseTagResolver()

        assert resolver.can_resolve(dep, project_config)
        assert resolver.resolve(dep, project_config) == (
            "tag_sha",
            PackageNamespaceVersionDependency(
                namespace="ccitestdep",
                version="2.1 Beta 1",
                package_name="CumulusCI-Test-Dep",
            ),
        )

    def test_no_managed_release(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/UnmanagedRepo",  # This repo has no namespace
            tag="release/1.0",
            unmanaged=False,
        )
        resolver = GitHubTagResolver()

        assert resolver.resolve(dep, project_config) == (
            "tag_sha",
            None,
        )

    def test_not_found(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/NoReleasesRepo"
        )
        resolver = GitHubReleaseTagResolver()

        assert resolver.can_resolve(dep, project_config)
        assert resolver.resolve(dep, project_config) == (None, None)


class TestGitHubUnmanagedHeadResolver:
    def test_unmanaged_head_resolver(self, project_config):
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/ReleasesRepo"
        )
        resolver = GitHubUnmanagedHeadResolver()

        assert resolver.can_resolve(dep, project_config)

        assert resolver.resolve(dep, project_config) == ("commit_sha", None)


class ConcreteGitHubReleaseBranchResolver(GitHubReleaseBranchResolver):
    def resolve(
        self, dep: GitHubDynamicDependency, context: BaseProjectConfig
    ) -> Tuple[Optional[str], Optional[StaticDependency]]:
        return (None, None)


class TestGitHubReleaseBranchResolver:
    def test_is_valid_repo_context(self):
        pc = BaseProjectConfig(UniversalConfig())

        pc.repo_info["branch"] = "feature/232__test"
        pc.project__git["prefix_feature"] = "feature/"

        assert ConcreteGitHubReleaseBranchResolver().is_valid_repo_context(pc)

        pc.repo_info["branch"] = "main"
        assert not ConcreteGitHubReleaseBranchResolver().is_valid_repo_context(pc)

    def test_can_resolve(self):
        pc = BaseProjectConfig(UniversalConfig())

        pc.repo_info["branch"] = "feature/232__test"
        pc.project__git["prefix_feature"] = "feature/"

        gh = ConcreteGitHubReleaseBranchResolver()

        assert gh.can_resolve(
            GitHubDynamicDependency(github="https://github.com/SFDO-Tooling/Test"),
            pc,
        )

        assert not gh.can_resolve(ConcreteDynamicDependency(), pc)

    def test_get_release_id(self):
        pc = BaseProjectConfig(UniversalConfig())
        pc.repo_info["branch"] = "feature/232__test"
        pc.project__git["prefix_feature"] = "feature/"

        assert ConcreteGitHubReleaseBranchResolver().get_release_id(pc) == 232

    def test_get_release_id__not_release_branch(self):
        pc = BaseProjectConfig(UniversalConfig())
        with mock.patch.object(
            BaseProjectConfig, "repo_branch", new_callable=mock.PropertyMock
        ) as repo_branch:
            repo_branch.return_value = None

            with pytest.raises(DependencyResolutionError) as e:
                ConcreteGitHubReleaseBranchResolver().get_release_id(pc)

            assert "Cannot get current branch" in str(e)

    def test_get_release_id__no_git_data(self):
        pc = BaseProjectConfig(UniversalConfig())
        with mock.patch.object(
            BaseProjectConfig, "repo_branch", new_callable=mock.PropertyMock
        ) as repo_branch:
            repo_branch.return_value = "feature/test"
            pc.project__git["prefix_feature"] = "feature/"

            with pytest.raises(DependencyResolutionError) as e:
                ConcreteGitHubReleaseBranchResolver().get_release_id(pc)

            assert "Cannot get current release identifier" in str(e)

    def test_locate_commit_status_package_id__not_found_with_parent(self, github):
        repo = github.repository("SFDO-Tooling", "TwoGPMissingRepo")
        assert ConcreteGitHubReleaseBranchResolver().locate_commit_status_package_id(
            repo, repo.branch("feature/232"), "Build Feature Test Package"
        ) == (None, None)


class TestGitHubReleaseBranchCommitStatusResolver:
    def test_2gp_release_branch_resolver(self, project_config):
        project_config.repo_branch = "feature/232__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPRepo"
        )

        assert resolver.can_resolve(dep, project_config)

        assert resolver.resolve(dep, project_config) == (
            "parent_sha",
            PackageVersionIdDependency(
                version_id="04t000000000000", package_name="CumulusCI-2GP-Test"
            ),
        )

    def test_commit_status_not_found(self, project_config):
        project_config.repo_branch = "feature/232__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPMissingRepo"
        )

        assert resolver.can_resolve(dep, project_config)
        assert resolver.resolve(dep, project_config) == (None, None)

    def test_repo_not_found(self, project_config):
        project_config.repo_branch = "feature/232__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/NonexistentRepo"
        )

        project_config.get_repo_from_url = mock.Mock(return_value=None)

        with pytest.raises(DependencyResolutionError) as exc:
            resolver.resolve(dep, project_config)

        assert "Unable to access GitHub" in str(exc)

    @mock.patch("cumulusci.core.dependencies.resolvers.find_repo_feature_prefix")
    def test_unable_locate_feature_prefix(
        self, find_repo_feature_prefix_mock, project_config
    ):
        find_repo_feature_prefix_mock.side_effect = NotFoundError
        project_config.repo_branch = "feature/232__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPRepo"
        )
        assert resolver.resolve(dep, project_config) == (None, None)

    def test_branch_not_found(self, project_config):
        project_config.repo_branch = "feature/290__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPRepo"
        )

        assert resolver.resolve(dep, project_config) == (None, None)


class TestGitHubReleaseBranchExactMatchCommitStatusResolver:
    def test_exact_branch_resolver(self, project_config):
        project_config.repo_branch = "feature/232__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchExactMatchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPRepo"
        )

        assert resolver.can_resolve(dep, project_config)

        assert resolver.resolve(dep, project_config) == (
            "feature/232__test_sha",
            PackageVersionIdDependency(
                version_id="04t000000000001", package_name="CumulusCI-2GP-Test"
            ),
        )

    def test_repo_not_found(self, project_config):
        project_config.repo_branch = "feature/232__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchExactMatchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/NonexistentRepo"
        )

        project_config.get_repo_from_url = mock.Mock(return_value=None)

        with pytest.raises(DependencyResolutionError) as exc:
            resolver.resolve(dep, project_config)

        assert "Unable to access GitHub" in str(exc)

    @mock.patch("cumulusci.core.dependencies.resolvers.find_repo_feature_prefix")
    def test_unable_locate_feature_prefix(
        self, find_repo_feature_prefix_mock, project_config
    ):
        find_repo_feature_prefix_mock.side_effect = NotFoundError
        project_config.repo_branch = "feature/232__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchExactMatchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPRepo"
        )
        assert resolver.resolve(dep, project_config) == (None, None)

    def test_branch_not_found(self, project_config):
        project_config.repo_branch = "feature/290__test"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchExactMatchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPRepo"
        )

        assert resolver.resolve(dep, project_config) == (None, None)

    def test_commit_status_not_found(self, project_config):
        project_config.repo_branch = "feature/232"
        project_config.project__git__prefix_feature = "feature/"

        resolver = GitHubReleaseBranchExactMatchCommitStatusResolver()
        dep = GitHubDynamicDependency(
            github="https://github.com/SFDO-Tooling/TwoGPMissingRepo"
        )

        assert resolver.can_resolve(dep, project_config)
        assert resolver.resolve(dep, project_config) == (None, None)


class TestResolverAccess:
    def test_get_resolver(self):
        assert isinstance(
            get_resolver(
                DependencyResolutionStrategy.STATIC_TAG_REFERENCE,
                GitHubDynamicDependency(github="https://github.com/SFDO-Tooling/Test"),
            ),
            GitHubTagResolver,
        )

    def test_get_resolver_stack__indirect(self):
        pc = BaseProjectConfig(UniversalConfig())

        strategy = get_resolver_stack(pc, "production")
        assert DependencyResolutionStrategy.RELEASE_TAG in strategy
        assert DependencyResolutionStrategy.BETA_RELEASE_TAG not in strategy

    def test_get_resolver_stack__customized_indirect(self):
        pc = BaseProjectConfig(UniversalConfig())

        pc.project__dependency_resolutions["preproduction"] = "include_beta"
        strategy = get_resolver_stack(pc, "preproduction")
        assert DependencyResolutionStrategy.BETA_RELEASE_TAG in strategy

    def test_get_resolver_stack__direct(self):
        pc = BaseProjectConfig(UniversalConfig())

        strategy = get_resolver_stack(pc, "commit_status")
        assert DependencyResolutionStrategy.COMMIT_STATUS_RELEASE_BRANCH in strategy

    def test_get_resolver_stack__fail(self):
        pc = BaseProjectConfig(UniversalConfig())

        with pytest.raises(CumulusCIException) as e:
            get_resolver_stack(pc, "bogus")

        assert "not found" in str(e)


class TestStaticDependencyResolution:
    def test_get_static_dependencies(self, project_config):
        gh = GitHubDynamicDependency(github="https://github.com/SFDO-Tooling/RootRepo")

        assert get_static_dependencies(
            project_config,
            dependencies=[gh],
            strategies=[DependencyResolutionStrategy.RELEASE_TAG],
        ) == [
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/DependencyRepo",
                subfolder="unpackaged/pre/top",
                unmanaged=True,
                ref="tag_sha",
            ),
            PackageNamespaceVersionDependency(
                namespace="foo",
                version="1.1",
                package_name="DependencyRepo",
                password_env_name="DEP_PW",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/DependencyRepo",
                subfolder="unpackaged/post/top",
                unmanaged=False,
                ref="tag_sha",
                namespace_inject="foo",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/first",
                unmanaged=True,
                ref="tag_sha",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/second",
                unmanaged=True,
                ref="tag_sha",
            ),
            PackageNamespaceVersionDependency(
                namespace="bar", version="2.0", package_name="RootRepo"
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/post/first",
                unmanaged=False,
                ref="tag_sha",
                namespace_inject="bar",
            ),
        ]

    def test_get_static_dependencies__ignore_namespace(self, project_config):
        gh = GitHubDynamicDependency(github="https://github.com/SFDO-Tooling/RootRepo")

        assert get_static_dependencies(
            project_config,
            dependencies=[gh],
            strategies=[DependencyResolutionStrategy.RELEASE_TAG],
            filter_function=dependency_filter_ignore_deps([{"namespace": "foo"}]),
        ) == [
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/DependencyRepo",
                subfolder="unpackaged/pre/top",
                unmanaged=True,
                ref="tag_sha",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/DependencyRepo",
                subfolder="unpackaged/post/top",
                unmanaged=False,
                ref="tag_sha",
                namespace_inject="foo",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/first",
                unmanaged=True,
                ref="tag_sha",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/second",
                unmanaged=True,
                ref="tag_sha",
            ),
            PackageNamespaceVersionDependency(
                namespace="bar", version="2.0", package_name="RootRepo"
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/post/first",
                unmanaged=False,
                ref="tag_sha",
                namespace_inject="bar",
            ),
        ]

    def test_get_static_dependencies__ignore_github(self, project_config):
        gh = GitHubDynamicDependency(github="https://github.com/SFDO-Tooling/RootRepo")

        assert get_static_dependencies(
            project_config,
            dependencies=[gh],
            strategies=[DependencyResolutionStrategy.RELEASE_TAG],
            filter_function=dependency_filter_ignore_deps(
                [{"github": "https://github.com/SFDO-Tooling/DependencyRepo"}]
            ),
        ) == [
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/first",
                unmanaged=True,
                ref="tag_sha",
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/pre/second",
                unmanaged=True,
                ref="tag_sha",
            ),
            PackageNamespaceVersionDependency(
                namespace="bar", version="2.0", package_name="RootRepo"
            ),
            UnmanagedGitHubRefDependency(
                github="https://github.com/SFDO-Tooling/RootRepo",
                subfolder="unpackaged/post/first",
                unmanaged=False,
                ref="tag_sha",
                namespace_inject="bar",
            ),
        ]



"""
Robot Debugger

"""

from cumulusci.tasks.robotframework.debugger import (
    Breakpoint,
    DebuggerCli,
    Keyword,
    Suite,
    Testcase,
)


class DebugListener(object):
    """A robot framework listener for debugging test cases

    This acts as the controller for the debugger. It is responsible
    for managing breakpoints, and pausing execution of a test when a
    breakpoint is hit.

    The listener is also responsible for instantiating the debugger UI
    (class DebuggerCli).

    """

    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, *breakpoints):
        self.show_initial_help = True
        self.current_file = None
        self.stack = []
        self.rdb = DebuggerCli(listener=self)
        if breakpoints:
            self.breakpoints = list(breakpoints)
        else:
            self.breakpoints = [
                Breakpoint(Keyword, "*::cumulusci.robotframework.Salesforce.Breakpoint")
            ]

    def start_suite(self, name, attrs):
        self.stack.append(Suite(name, attrs))

    def start_test(self, name, attrs):
        self.stack.append(Testcase(name, attrs))

    def start_keyword(self, name, attrs):
        context = Keyword(name, attrs)
        self.stack.append(context)
        self.break_if_breakpoint()

    def end_keyword(self, name, attrs):
        self.stack.pop()

    def end_test(self, name, attrs):
        self.stack.pop()

    def end_suite(self, name, attrs):
        self.stack.pop()

    def do_step(self):
        """Single-step through the code

        This will set a temporary breakpoint on the next keyword in
        the current context before continuing. Once the breakpoint
        is hit, it will be removed from the list of breakpoints.
        """
        breakpoint = Breakpoint(
            Keyword, "{}::*".format(self.stack[-2].longname), temporary=True
        )
        self.breakpoints.append(breakpoint)

    def break_if_breakpoint(self):
        """Pause test execution and issue a prompt if we are at a breakpoint"""

        # filter breakpoints to only those that match the current context
        # (eg: Suite, Testcase, Keyword), and iterate over them looking
        # for a match.
        for breakpoint in [
            bp
            for bp in self.breakpoints
            if isinstance(self.stack[-1], bp.breakpoint_type)
        ]:
            statement = "{}::{}".format(self.stack[-2].longname, self.stack[-1].name)
            if breakpoint.match(statement):
                if breakpoint.temporary:
                    self.breakpoints.remove(breakpoint)

                intro = "\n"
                if self.show_initial_help:
                    self.show_initial_help = False
                    intro += self.rdb.intro
                intro += "\n> {}\n-> {}".format(
                    self.stack[-2].longname, str(self.stack[-1])
                )

                # Note: this call won't return until a debugger command
                # has been issued which returns True (eg: 'continue' or 'step')
                self.rdb.cmdloop(intro)
                return
